r"""This package contains the model base class and some implemented models."""

__all__ = [
    "BaseModel",
    "VanillaModel",
]

from gravitorch.models.base import BaseModel
from gravitorch.models.vanilla import VanillaModel
